# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
[TODO: write this}
"""

import asyncore
import asynchat
import cPickle
import logging
import socket
import time
import threading

from duo.protocols import PlayerUI
from duo.protocols import PlayerBackend
from duo.protocols import ServerUI
from duo.protocols import Message
from duo.protocols import Option

TERMINATOR = '\r\n'

OPTIONS = [
    Option('create_server', False, 'Create server'),
    Option('join_server', True, 'Join server'),
    Option('hostname', 'localhost', 'Hostname'),
    Option('port', 8881, 'Port')
]

class PlayerBackend(object):
    """The player class from the referee point of view."""
    def __init__(self):
        pass

    def getName(self):
        """Retrieve the name of this player."""
        raise NotImplementedError, \
              '%s shall define getName' % self.__class__.__name__

    def sendMessage(self, aBundle):
        """Send a message to this player.

        The bundle argument is a list of massages.
        This method is used by the referee any time it shall send some
        piece of information to the player representation (See Referee).
        """
        raise NotImplementedError, \
              '%s shall define sendMessage' % self.__class__.__name__

    def isReady(self):
        """Check if this player is ready."""
        raise NotImplementedError, \
              '%s shall implement isReady' % self.__class__.__name__


class ServerUI(object):
    """The server minimal interface.
    """
    def __init__(self, aNbPlayers):
        self.__referee = referee.Referee(aNbPlayers)


    def getReferee(self):
        """Retrieve the referee of the current game provided by this server."""
        return self.__referee


    def mainloop(self):
        raise NotImplementedError, \
              '%s shall implement mainloop' % self.__class__.__name__


    def getCoordinates(self):
        raise NotImplementedError, \
              '%s shall implement getCoordinates' % self.__class__.__name__



class Player(threading.Thread, asynchat.async_chat, PlayerUI):
    CONNECTED = 'connected'
    AUTHENTICATED = 'authenticated'
    PLAYING = 'playing'

    def __init__(self, aName, aOptions, aIsRobot):
        threading.Thread.__init__(self, name = aName)
        asynchat.async_chat.__init__(self)
        PlayerUI.__init__(self, aName, aOptions, aIsRobot)

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect((self.getOption('hostname'), self.getOption('port')))
        self.set_terminator(TERMINATOR)

        self.register(Message.AUTH, self.onAuthCb)
        self.register(Message.DEAL, self.onDealCb)

        self.__data = []
        self.__status = Player.CONNECTED

        self.start()


    def run(self):
        asyncore.loop()


    def authenticate(self):
        self.sendToServer(Message(Message.HELO, self.getName()))


    def onAuthCb(self, aMessage, aArgs):
        if self.__status != Player.PLAYING:
            self.__status = Player.AUTHENTICATED
            self.sendToServer(Message(Message.READY, self.getName())) #TODO: move out


    def onDealCb(self, aMessage, aArgs):
        if self.__status != Player.CONNECTED:
            self.__status = self.PLAYING
            self._PlayerUI__hand = aMessage.getPayload()


    def move(self, aCard, aColor = 'N'):
        """Play a card."""
        if aCard in self.getHand():
            self.sendToServer(Message(Message.PLAY, (aCard, aColor)))
        else:
            print 'TODO: ERROR...'


    def draw(self):
        """Draw a given number of cards."""
        self.sendToServer(Message(Message.DRAW, None))


    def passTurn(self):
        """Pass the turn."""
        self.sendToServer(Message(Message.PASS, None))


    def sendToServer(self, aBundle):
        self.push(cPickle.dumps(aBundle) + TERMINATOR)


    def collect_incoming_data(self, aData):
        logging.getLogger('duo').debug(
            'Received %d bytes from %s' % (len(aData), self.getName()))
        self.__data.append(aData)


    def found_terminator(self):
        lMessage = ''.join(self.__data).strip()
        try:
            self.parse(cPickle.loads(lMessage))

        except Exception, e:
            logging.getLogger('duo').error('parse error: %s' % (e))

        self.__data = []


    def handle_connect(self):
        logging.getLogger('duo').info(str(self.getName()) + ' connected.')


    def handle_close(self):
        logging.getLogger('duo').info(str(self.getName()) + ' disconnected.')
        self.close()



class PlayerSock(asynchat.async_chat, PlayerBackend):
    CONNECTED = 'connected'
    AUTHENTICATED = 'authenticated'

    def __init__(self, aReferee, *aArgs):
        asynchat.async_chat.__init__(self, *aArgs)
        PlayerBackend.__init__(self)

        self.set_terminator(TERMINATOR)

        self.__data = []
        self.__user = None
        self.__ready = False
        self.__referee = aReferee
        self.__status = PlayerSock.CONNECTED


    def getName(self):
        return self.__user


    def sendMessage(self, aMessageBundle):
        # TODO: rename to sendMessageBundle ?
        #        logging.getLogger('duo').debug('sending %s to %s' %
        #                                       (aMessage.getType(), self.__user))
        self.push(cPickle.dumps(aMessageBundle) + TERMINATOR)


    def isReady(self):
        return self.__ready


    def authenticate(self, aUserName):
        logging.getLogger('duo').info('authenticating %s' % (self.__user))
        if self.__status == self.CONNECTED:
            self.__user = aUserName
            self.__status = PlayerSock.AUTHENTICATED
            self.__referee.addPlayer(self)
        else:
            print 'TODO: ERROR...'


    def setReady(self):
        logging.getLogger('duo').info('%s is ready' % (self.__user))
        if self.__status == self.AUTHENTICATED:
            self.__ready = True
            self.__referee.startGame()
        else:
            print 'TODO: ERROR...'


    def move(self, aCard, aColor):
        logging.getLogger('duo').debug(
            '%s plays %s (%s)' % (self.__user, aCard, aColor))
        if self.__status == self.AUTHENTICATED:
            self.__referee.move(self, aCard, aColor)
        else:
            print 'TODO: ERROR...'


    def draw(self):
        logging.getLogger('duo').debug('%s draws card(s)' % self.__user)
        if self.__status == self.AUTHENTICATED:
            self.__referee.draw(self)
        else:
            print 'TODO: ERROR...'


    def passTurn(self):
        """Pass the turn."""
        logging.getLogger('duo').debug('%s passes' % self.__user)
        if self.__status == self.AUTHENTICATED:
            self.__referee.passTurn(self)
        else:
            print 'TODO: ERROR...'


    def collect_incoming_data(self, aData):
        logging.getLogger('duo').debug(
            'Received %d bytes from %s' % (len(aData), self.__user))
        self.__data.append(aData)


    def found_terminator(self):
        lFrame = ''.join(self.__data).strip()
        try:
            lMessage = cPickle.loads(lFrame)
            if lMessage.getType() == Message.HELO:
                self.authenticate(lMessage.getPayload())

            elif lMessage.getType() == Message.READY:
                self.setReady()

            elif lMessage.getType() == Message.PLAY:
                print lMessage.getPayload()
                lCard, lColor = lMessage.getPayload()
                self.move(lCard, lColor)

            elif lMessage.getType() == Message.DRAW:
                self.draw()

            elif lMessage.getType() == Message.PASS:
                self.passTurn()

            elif lRequest == '':
                pass

            else:
                raise Exception, 'unexpected %s\n' % lMessage.getType()

        except Exception, e:
            logging.getLogger('duo').error('exception %s' % (e))
            self.push(Message.ERROR + ' Exception: %s\n' % (e))
            self.close()

        self.__data = []


    def handle_close(self):
        logging.getLogger('duo').info(str(self.__user) + ' disconnected.')
        self.close()



class Server(ServerUI, asyncore.dispatcher):
    def __init__(self, aNbPlayers):
        ServerUI.__init__(self, aNbPlayers)
        asyncore.dispatcher.__init__(self)


    def start(self, aPort):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(('', aPort))
        self.listen(5)
        asyncore.loop()


    def mainloop(self):
        # TODO: get the port from somewhere
        self.start(8881)


    def getCoordinates(self):
        return ('', aPort)


    def handle_accept(self):
        """Accept and manage a new connection.
        """
        lSock, lAddress = self.accept()
        PlayerSock(self.getReferee(), lSock)
